import math
import pygame
from cat_royale.classes.item import Weapon, Tool
from cat_royale.classes.inventory import InventoryHUD, CraftingHUD
from cat_royale.classes.message_service import MessageService
from cat_royale.classes.functions import get_item_sprite_image, scale_sprites
from cat_royale.classes.inventory import Inventory
from cat_royale.classes.config import Config


class Character:
    """
    The Character class represents a character in the game, either the player or an enemy.

    Attributes:
        player (bool): True if the character is a player, False if it's an enemy.\n
        sprites (dict): Dictionary containing lists of sprites for different actions.\n
        hp (int): Health points of the character.\n
        max_hp (int): Maximum health points a character can have.\n
        hunger (int): Hunger points of the character.\n
        max_hunger (int): Maximum hunger points a character can have.\n
        doing_percentage (int): The percentage of progress made in doing an action.\n
        inventory (Inventory): The character's inventory.\n
        inventory_hud (InventoryHUD): The HUD (Heads Up Display) for the character's inventory.\n
        crafting_hud (CraftingHUD): The HUD for the character's crafting.\n
        idle_generator (generator): Generator for the idle sprites.\n
        next_sprite (pygame.Surface): The next sprite to be drawn for the character.\n
        wait_for_idle (int): Time to wait before returning to idle.\n
        walk_generator (generator): Generator for the walking sprites.\n
        walk_speed (int): Speed of the character animation when walking.\n
        die_generator (generator): Generator for the dying sprites.\n
        died_speed (int): Speed of the character when dying.\n
        doing_generator (generator): Generator for the doing action sprites.\n
        doingThing (bool): Whether the character is doing an action.\n
        done (bool): Whether the character has finished doing an action.\n
        doingTime (int): Time taken to do an action.\n
        do_attack (bool): Whether the character is performing an attack.\n
        attack_generator (generator): Generator for the attack sprites.\n
        delay (int): Delay in sprite animation.\n
        current_time (int): Current time in the game.\n
        position (tuple): The current position of the character on the screen.\n
        tick_count (int): The count of ticks in the game.\n
        onblock (Block): The block the character is currently on.\n
    """

    def __init__(self, player=True) -> None:
        """
        Constructs all the necessary attributes for the character object.

        Args:
            player (bool): True if the character is a player, False if it's an enemy. Defaults to True.
        """
        self.player = player

        self.sprites = {
            'walk': [],
            'died': [],
            'idle': [],
            'doing': [],
            "attack": []

        }

        self.hp = 100
        self.max_hp = 100

        self.hunger = 100
        self.max_hunger = 100

        self.doing_percentage = 0

        self.inventory = Inventory()

        self.inventory_hud = InventoryHUD(self.inventory)
        self.inventory_hud.update_slots()

        self.crafting_hud = CraftingHUD(self)

        self.load_sprite()
        self.idle_generator = self.generate_idle()
        self.next_sprite = None
        self.wait_for_idle = 100

        self.walk_generator = self.generate_walk()
        self.walk_speed = 5

        self.die_generator = self.generate_died()
        self.died_speed = 10

        self.doing_generator = self.generate_doing()
        self.doingThing = False
        self.done = True
        self.doingTime = 100

        self.do_attack = False
        self.attack_generator = self.generate_attack()

        self.delay = Config.data["animation_delay"]
        self.current_time = 0

        self.position = (0, 0)

        self.tick_count = 0

        self.onblock = None

    def add_hp(self, hp):
        """
        Adds a specified amount of health points (hp) to the character, without exceeding the maximum HP.
        """
        self.hp += hp if self.hp + hp < self.max_hp else self.max_hp - self.hp

    def add_hunger(self, hunger):
        """
        Adds a specified amount of hunger points to the character, without exceeding the maximum hunger.
        """
        self.hunger += hunger if self.hunger + \
            hunger < self.max_hunger else self.max_hunger - self.hunger

    def tick(self):
        """
        Handles the tick event for the character, which updates the hunger and HP values over time.
        """
        self.tick_count += 1
        if self.tick_count > 60:
            self.tick_count = 0
            self.hunger += Config.data["hunger_per_tick"] if self.hunger > 0 else 0
            if self.hunger <= 0:
                if self.hp > 10:
                    self.hp += Config.data["health_if_hungry"]

    def add_event_to_craft_hud(self, event):
        """
        Adds an event to the crafting HUD's event box.
        """
        self.crafting_hud.add_event_to_box(event)

    def load_sprite(self):
        """
        Loads the sprite images for the character from the configuration data.
        """
        cat_config = "cat_sprite_indexes" if self.player else "enemy_cat_sprite_indexes"
        img = pygame.image.load(
            Config.images["cat" if self.player else "enemy_cat"]).convert_alpha()

        walk = Config.data[cat_config]["walk"]
        self.sprites['walk'] = scale_sprites(get_item_sprite_image(
            img, walk["row"], walk["count"]), 1.5)

        died = Config.data[cat_config]["died"]
        self.sprites['died'] = scale_sprites(get_item_sprite_image(
            img, died['row'], died['count']), 1.5)

        idle = Config.data[cat_config]["idle"]
        self.sprites['idle'] = scale_sprites(get_item_sprite_image(
            img, idle['row'], idle['count']), 1.5)

        doing = Config.data[cat_config]["doing"]
        self.sprites['doing'] = scale_sprites(get_item_sprite_image(
            img, doing['row'], doing['count']), 1.5)

        attack = Config.data[cat_config]["attack"]
        self.sprites['attack'] = scale_sprites(get_item_sprite_image(
            img, attack['row'], attack['count']), 1.5)

    def generate_walk(self):
        """
        A generator that iterates through the character's walking sprites.
        """
        for sprite in self.sprites["walk"]:
            yield sprite

    def walk(self, direction):
        """
        Updates the character's sprite and position based on the given direction ('left', 'right', 'up', 'down').
        """
        if self.current_time < self.walk_speed:
            self.current_time += 1
            return
        try:
            self.next_sprite = next(self.walk_generator)

        except StopIteration:
            self.walk_generator = self.generate_walk()
            self.next_sprite = next(self.walk_generator)
        if direction == 'left':
            self.next_sprite = pygame.transform.flip(
                self.next_sprite, True, False)
        self.current_time = 0

    def died(self, screen):
        """
        Handles the character's death by updating the sprite and displaying it on the screen.
        """
        if self.next_sprite is None:
            self.next_sprite = next(self.die_generator)

        if self.current_time < self.died_speed:
            self.current_time += 1
            screen.blit(self.next_sprite, self.position)
            return
        try:
            self.next_sprite = next(self.die_generator)

        except StopIteration:
            self.die_generator = self.generate_died()
            self.next_sprite = next(self.die_generator)

        screen.blit(self.next_sprite, self.position)
        self.current_time = 0

    def generate_died(self):
        """
        A generator that iterates through the character's dying sprites.
        """
        for sprite in self.sprites["died"]:
            yield sprite

    def attack(self, screen):
        """
        Handles the character's attack by updating the sprite and displaying it on the screen.
        """
        if self.next_sprite is None:
            self.next_sprite = next(self.attack_generator)

        if self.current_time < self.died_speed:
            self.current_time += 1
            screen.blit(self.next_sprite, self.position)
            return
        try:
            self.next_sprite = next(self.attack_generator)

        except StopIteration:
            self.attack_generator = self.generate_attack()
            self.next_sprite = next(self.attack_generator)
            self.do_attack = False

        screen.blit(self.next_sprite, self.position)
        self.current_time = 0

    def generate_attack(self):
        """
        A generator that iterates through the character's attack sprites.
        """
        for sprite in self.sprites["attack"]:
            yield sprite

    def use_weapon_to_attack(self):
        """
        Uses the selected weapon in the inventory to attack nearby characters.
        """
        selected_item = self.inventory.slots[self.inventory_hud.selected_slot]
        if not self.onblock:
            return

        coords = self.onblock.indexes
        to_be_checked = [(coords[0]-1, coords[1]-1), (coords[0]-1, coords[1]), (coords[0]-1, coords[1]+1), (coords[0], coords[1]-1),
                         (coords[0], coords[1]+1), (coords[0]+1, coords[1]-1), (coords[0]+1, coords[1]), (coords[0]+1, coords[1]+1), coords]
        characters_to_be_damaged = []
        for character in Config.all_characters:
            if character.get_index_coords() in to_be_checked:
                characters_to_be_damaged.append(character)
        if isinstance(selected_item, Weapon):
            for character in characters_to_be_damaged:
                if character != self:
                    selected_item.use(character)

            if selected_item.durability == 0:
                self.inventory.slots[self.inventory_hud.selected_slot] = None
        else:
            for character in characters_to_be_damaged:
                if character != self:
                    character.hp -= Config.data["damage"]['FIST']

    def do_damage(self, damage):
        """
        Reduces the character's HP by the specified amount of damage.
        """
        self.hp -= damage if damage < self.hp else self.hp

    def get_index_coords(self):
        """
        Returns the index coordinates of the block the character is currently on.
        """
        return self.onblock.indexes

    def generate_idle(self):
        """
        A generator that iterates through the character's idle sprites.
        """
        for sprite in self.sprites["idle"]:
            yield sprite

    def idle(self, screen):
        """
        Updates the character's sprite to an idle state and displays it on the screen.
        """
        if self.next_sprite is None:
            self.next_sprite = next(self.idle_generator)

        if self.current_time < self.delay:
            self.current_time += 1
            screen.blit(self.next_sprite, self.position)
            return
        try:
            self.next_sprite = next(self.idle_generator)

        except StopIteration:
            self.idle_generator = self.generate_idle()
            self.next_sprite = next(self.idle_generator)

        screen.blit(self.next_sprite, self.position)
        self.current_time = 0

    def generate_doing(self):
        """
        A generator that iterates through the character's 'doing' (action) sprites.
        """
        for sprite in self.sprites["doing"]:
            yield sprite

    def doing(self, screen: pygame.Surface):
        """
        Updates the character's sprite to a 'doing' (action) state and displays it on the screen.
        """
        if self.next_sprite is None:
            self.next_sprite = next(self.doing_generator)

        if self.current_time < self.delay:
            self.current_time += 1
            self.doing_percentage += 1
            if self.doing_percentage > self.next_sprite.get_width():
                self.doing_percentage = 0
            screen.blit(self.next_sprite, self.position)
            return
        try:
            self.next_sprite = next(self.doing_generator)

        except StopIteration:
            self.doing_generator = self.generate_doing()
            self.next_sprite = next(self.doing_generator)

        screen.blit(self.next_sprite, self.position)
        self.current_time = 0

    def doing_bar(self):
        """
        Returns a pygame Surface representing the progress bar for the character's current action.
        """
        bar_surface = pygame.Surface((self.doing_percentage, 1))
        bar_surface.fill((255, 255, 255))
        return bar_surface

    def draw(self, screen: pygame.Surface):
        """
        Draws the character's current sprite on the screen, handling different states and actions.
        """
        if self.onblock and len(self.onblock.items) > 0 and self.player:
            MessageService.add(
                {"text": "Press F to pickup", "severity": "info", "duration": 1})

        self.tick()
        if self.hp <= 0:
            self.died(screen)
            return

        self.wait_for_idle += 1

        if self.wait_for_idle >= 5:
            self.wait_for_idle = 100

            if self.doingThing and self.doingTime > 0:
                if not self.done:
                    self.done = True
                self.doing(screen=screen)

                screen.blit(self.doing_bar(),
                            (self.position[0], self.position[1] + self.next_sprite.get_height()+10))
                self.doingTime -= 1

            elif self.do_attack:
                self.attack(screen)

            else:
                self.idle(screen)

        else:
            self.check_wall_boundries(screen)
            screen.blit(self.next_sprite, self.position)

        self.pickup_time()

    def move(self, direction):
        """
        Moves the character in the specified direction ('left', 'right', 'up', 'down'), taking into account the character's HP and hunger.
        """
        if self.hp <= 0:
            return
        if self.doingThing:
            return

        self.wait_for_idle = 0
        self.walk(direction)
        speed_modifier = 4
        if self.hp < 50 and self.hp >= 30:
            speed_modifier = 0.9
        elif self.hp < 30 and self.hp >= 10:
            speed_modifier = 0.75
        elif self.hp < 10:
            speed_modifier = 0.6

        if self.hunger < 50 and self.hunger >= 20:
            speed_modifier *= 0.9
        elif self.hunger < 20 and self.hunger > 0:
            speed_modifier *= 0.8
        elif self.hunger <= 0:
            speed_modifier *= 0.5

        if direction == 'left':
            self.position = (self.position[0] -
                             1*speed_modifier, self.position[1])
        elif direction == 'up':
            self.position = (self.position[0],
                             self.position[1] - 1*speed_modifier)
        elif direction == 'down':
            self.position = (self.position[0],
                             self.position[1] + 1*speed_modifier)
        elif direction == 'right':
            self.position = (self.position[0] +
                             1*speed_modifier, self.position[1])

    def check_wall_boundries(self, screen):
        """
        Ensures the character does not move beyond the screen boundaries.
        """
        if self.position[0] < 0:
            self.position = (0, self.position[1])
        elif self.position[0] > screen.get_width() - self.next_sprite.get_width():

            self.position = (screen.get_width() -
                             self.next_sprite.get_width(), self.position[1])

        if self.position[1] < 0:
            self.position = (self.position[0], 0)

    def get_position(self):
        """
        Returns the current position of the character as a tuple (x, y).
        """
        return (self.position[0] + self.sprites['idle'][0].get_width() // 2, self.position[1] + self.sprites['idle'][0].get_height() // 2)

    def get_health(self):
        """
        Returns the character's current health points (HP, Hunger) as an integer.
        """
        return (self.hp, self.hunger)

    def set_onblock(self, block):
        """
        Sets the character's current block, updating the character's position to be centered on the specified block.
        """
        self.onblock = block

    def pickup(self):
        """
        Attempt to pick up an item from the block the character is standing on.

        Returns:
            bool: False if the item could not be picked up, True otherwise.
        """
        if self.onblock and len(self.onblock.items) > 0:
            if not self.check_necessary_item_for_pickup(self.onblock.items[-1].type):
                if self.player:
                    MessageService.add(
                        {'text': "Don't have necessary equipment!", "severity": "warning"})
                return False
            if self.inventory.isFull:
                if self.player:
                    MessageService.add(
                        {'text': "Inventory is full", "severity": "warning"})
                return False
            self.doingThing = True
            try:
                self.doingTime = Config.data["item_pickup_time"][self.onblock.items[0].type]
            except KeyError:
                self.doingTime = 1

        return False

    def check_necessary_item_for_pickup(self, pickup):
        """
        Checks if the character has the necessary item in their inventory to pick up the specified item.

        Returns True if the necessary item is found, otherwise False.
        """

        if pickup in Config.data["necessary_items_for_pickup"]:
            necessary = Config.data["necessary_items_for_pickup"][pickup]
            for (key, item) in self.inventory.slots.items():
                if item and item.type == necessary:
                    item.use()

                    if isinstance(item, Weapon) or isinstance(item, Tool) and item.durability <= 0:
                        self.inventory.slots[key] = None
                    return True
            return False
        return True

    def use_slot(self):
        """
        Selects the specified inventory slot (slot_number) for use. If an item is in the selected slot, the character will use it.
        """
        if self.inventory.slots[self.inventory_hud.selected_slot]:
            self.inventory.slots[self.inventory_hud.selected_slot].use(self)

    def pickup_time(self):
        """
        Picks up an item from the block the character is standing on,
        if the character is currently picking up an item, and the necessary time has passed.
        """
        if not self.done:
            return
        if self.done and self.doingTime <= 0:
            picked_up = self.onblock.remove_item_from_top()
            if picked_up:
                self.inventory.add_item_to_stack(picked_up)
                self.onblock.draw(Config.map_layer)
                self.inventory_hud.update_slots()
            self.done = False
            self.doingThing = False
            self.doingTime = 0
            self.doing_percentage = 0


class Enemy(Character):
    """
    Enemy class represents an enemy character in the game, derived from the Character class.
    """

    def __init__(self) -> None:
        """
        Initializes an instance of the Enemy class with specific attributes and calls the superclass constructor.
        """
        super().__init__(player=False)
        self.wait = 10
        self.picking = False

    def ai(self):
        """
        Controls the artificial intelligence of the enemy, including movement, attacking, and eating behaviors.

        Args:
            map_layer: The game's map layer.

        """
        if self.hp <= 0:
            return
        if not self.onblock:
            return
        coords = self.onblock.indexes
        to_be_checked = [coords,
                         (coords[0] - 1, coords[1] - 1),
                         (coords[0] - 1, coords[1]),
                         (coords[0] - 1, coords[1] + 1),
                         (coords[0], coords[1] - 1),
                         (coords[0], coords[1] + 1),
                         (coords[0] + 1, coords[1] - 1),
                         (coords[0] + 1, coords[1]),
                         (coords[0] + 1, coords[1] + 1),
                         (coords[0] - 2), (coords[1] - 2),
                         (coords[0] - 2, coords[1] - 1),
                         (coords[0] - 2, coords[1]),
                         (coords[0] - 2, coords[1] + 1),
                         (coords[0] - 2, coords[1] + 2),
                         (coords[0] - 1, coords[1] - 2),
                         (coords[0] - 1, coords[1] + 2),
                         (coords[0], coords[1] - 2),
                         (coords[0], coords[1] + 2),
                         (coords[0] + 1, coords[1] - 2),
                         (coords[0] + 1, coords[1] + 2),
                         (coords[0] + 2, coords[1] - 2),
                         (coords[0] + 2, coords[1] - 1),
                         (coords[0] + 2, coords[1]),
                         (coords[0] + 2, coords[1] + 1),
                         (coords[0] + 2, coords[1] + 2)]  # 5x5 area around the character
        characters_to_be_damaged = []
        for character in Config.all_characters:
            if character.get_index_coords() in to_be_checked:
                characters_to_be_damaged.append(character)

        if len(characters_to_be_damaged) > 1:
            has_alive = [
                character for character in characters_to_be_damaged if character.hp > 0]
            if len(has_alive) > 0:
                if self.wait <= 0:
                    for character in characters_to_be_damaged:
                        if character != self and character.hp > 0:
                            self.auto_move_to_pos(character.position)
                            self.do_attack = True
                            self.use_weapon_to_attack()
                    self.wait = 10
                else:
                    self.wait -= 1
                return
        if self.hunger < 30:
            self.search_food()

    def search_food(self):
        closest_food = None
        closest_distance = 100000
        for food in Config.items['FOOD']:
            if food.type == "MUSHROOM":
                return
            food_distance = math.sqrt(
                (food.coords[0]-self.position[0])**2 + (food.coords[1]-self.position[1])**2)
            if food_distance < closest_distance:
                closest_distance = food_distance
                closest_food = food
        if closest_food:
            self.auto_move_to_pos(closest_food.coords)
            if closest_distance < 2:
                self.pickup()

                self.eat()

    def eat(self):
        """
        Makes the enemy eat an item from its inventory to restore hunger.
        """
        if self.inventory.slots[0]:
            self.inventory.slots[0].use(self)

    def auto_move_to_pos(self, pos):
        """
        Automatically moves the enemy towards the specified position.

        Args:
            pos: A tuple representing the target position (x, y).
        """
        if self.hp <= 0:
            return

        if self.position[0] < pos[0]:
            self.move('right')
        elif self.position[0] > pos[0]:
            self.move('left')
        if self.position[1] < pos[1]:
            self.move('down')
        if self.position[1] > pos[1]:
            self.move('up')

    def draw(self, screen):
        """
        Draws the enemy's current sprite on the screen, including the health bar.

        Args:
            screen: A pygame.Surface representing the game screen.
        """
        super().draw(screen)
        pygame.draw.rect(self.next_sprite, (90, 0, 0), (0, self.next_sprite.get_height(
        )-2, self.next_sprite.get_width(), 2))
        pygame.draw.rect(self.next_sprite, (255, 0, 0), (0, self.next_sprite.get_height(
        )-2, self.next_sprite.get_width()*(self.hp/100), 2))



from enum import Enum

import pygame
from cat_royale.classes.map_reader import get_map_sprite_image
from cat_royale.classes.message_service import MessageService


class MAPCOLOR(Enum):
    """Item enum

    colorcode of the items on the map
    """
    GRASS = '#32C832'
    WOOD = "#C86432"
    STONE = "#C8C8C8"
    STICK = "#F0B478"
    BERRY = "#FF0000"
    CARROT = "#FAC800"

    def rgb(self, hexa) -> tuple:
        """Converts a hexa colorcode to a rgb tuple

        Args:
            hexa (str): hexa colorcode

        Returns:
            tuple: rgb tuple
        """
        hexa = int(hexa[1:], 16)
        return ((hexa >> 16) & 0xFF, (hexa >> 8) & 0xFF, hexa & 0xFF)


class ITEM(Enum):
    """Item position on sprite sheet

    type of the items
    """
    WOOD = (17, 0)
    CARROT = (14, 6)
    BERRY = (14, 4)
    STICK = (12, 2)
    STONE = (17, 1)
    AXE = (4, 6)
    PICKAXE = (4, 5)
    TORCH = (10, 10)
    WOOD_SWORD = (5, 0)
    STONE_SWORD = (5, 1)
    MUSHROOM = (14, 12)
    APPLE = (14, 0)


def load_item_image(item_type: ITEM) -> pygame.Surface:
    """Loads the image of the item

    Args:
        item_type (ITEM): type of the item

    Returns:
        pygame.Surface: image of the item
    """
    sprite_sheet = pygame.image.load(
        'cat_royale/assets/images/items.png').convert_alpha()

    return get_map_sprite_image(sprite_sheet, ITEM[item_type].value)


class Item:
    def __init__(self, item_image, item_type: ITEM = None, stack_size: int = 1) -> None:
        self.count = 1
        self.max = stack_size
        self.item_image = item_image.copy()
        self.type = item_type

    def __str__(self) -> str:
        return self.type.name + " " + str(self.count)

    def use(self, *args):
        pass


class Food(Item):
    def __init__(self, item_image, item_type: ITEM = None, stack_size: int = 1, hunger: int = 0, health: int = 0) -> None:
        super().__init__(item_image.copy(), item_type, stack_size)
        self.hunger = hunger
        self.health = health

    def use(self, character):
        if self.count > 0:
            if character.player:
                MessageService.add(
                    {"text": "You ate the " + ITEM[self.type].name.lower() + ".", "severity": "info", "duration": 100})
            character.inventory.subtract_item(self.type, 1)
            character.inventory_hud.update_slots()
            character.add_hp(self.health)
            character.add_hunger(self.hunger)


class Weapon(Item):
    def __init__(self, item_image, item_type: ITEM = None, stack_size: int = 1, damage: int = 0, durability: int = 0) -> None:
        super().__init__(item_image.copy(), item_type, stack_size)
        self.damage = damage
        self.durability = durability
        self.max_durability = durability

    def use(self, character):
        self.durability -= 1
        character.do_damage(self.damage)
        self.item_image.blit(self.make_durability_bar(),
                             (0, self.item_image.get_height() - 2))

    def make_durability_bar(self):

        durability_bar = pygame.Surface((self.item_image.get_width(), 2))
        durability_bar.fill((90, 0, 0))
        durability_bar.fill(
            (255, 0, 0), (0, 0, self.item_image.get_width() * self.durability / 100, 2))
        return durability_bar


class Material(Item):
    def __init__(self, item_image, item_type: ITEM = None, stack_size: int = 1) -> None:
        super().__init__(item_image.copy(), item_type, stack_size)


class Tool(Item):
    count = 0

    def __init__(self, item_image, item_type: ITEM = None, stack_size: int = 1, durability: int = 0) -> None:
        super().__init__(item_image.copy(), item_type, stack_size)
        self.durability = durability
        self.max_durability = durability
        Tool.count += 1
        self.id = Tool.count

    def use(self):
        self.durability -= 1
        self.item_image.blit(self.make_durability_bar(),
                             (0, self.item_image.get_height() - 2))

    def make_durability_bar(self):

        durability_bar = pygame.Surface((self.item_image.get_width(), 2))
        durability_bar.fill((90, 0, 0))
        durability_bar.fill(
            (255, 0, 0), (0, 0, self.item_image.get_width() * (self.durability / self.max_durability), 2))
        return durability_bar



import pygame
from cat_royale.classes.item import Weapon, Tool, Food, Material
from cat_royale.classes.box import Box
from cat_royale.classes.config import Config
from cat_royale.classes.functions import scale_image

from cat_royale.classes.item import load_item_image
from cat_royale.classes.text_display import TextDisplay


class Inventory:
    def __init__(self) -> None:
        self.slots = {x: None for x in range(10)}
        self.isFull = False

    def __str__(self) -> str:
        return str(self.slots)

    def add_item_to_stack(self, item):
        for slot in self.slots:
            if self.slots[slot] is not None and self.slots[slot].type == item.type and self.slots[slot].count < self.slots[slot].max:
                self.slots[slot].count += item.count
                self.isFull = self.check_full()
                return True

        for slot in self.slots:
            if self.slots[slot] is None:
                self.slots[slot] = item
                self.isFull = self.check_full()
                return True

        return False

    def check_full(self):
        for slot in self.slots:
            if not self.slots[slot]:
                return False

            if self.slots[slot].count < self.slots[slot].max:
                return False

        return True

    def subtract_item(self, item, count):
        for slot in self.slots:
            if self.slots[slot] is not None and self.slots[slot].type == item:
                self.slots[slot].count -= count
                if self.slots[slot].count <= 0:
                    self.slots[slot] = None
                return True
        return False

    def remove_item_from_slot(self, slot: int):
        removed_item = self.slots[slot]
        self.slots[slot] = None
        return removed_item

    def use_item(self, slot: int):
        if self.slots[slot] is not None:
            self.slots[slot].use()
            if self.slots[slot].count <= 0:
                self.slots[slot] = None

            return True
        return False


class InventoryHUD:
    def __init__(self, inventory: Inventory) -> None:
        self.inventory_offset = 12
        self.inv_bar = scale_image(
            pygame.image.load(Config.images["inventory_bar"]), 2)
        self.selected_slot_img = scale_image(pygame.image.load(
            Config.images["inventory_slot_selected"]), 2)

        self.inventory_Surface = pygame.Surface(
            (self.inv_bar.get_width(), self.inv_bar.get_height()))
        self.selected_slot = 0

        self.inventory = inventory
        self.update_slots()

    def draw(self, screen: pygame.Surface):
        screen.blit(self.inventory_Surface, (screen.get_width() // 2 - self.inventory_Surface.get_width() // 2,
                                             screen.get_height() - self.inventory_Surface.get_height()))

    def select_slot(self, slot: int):
        if slot > 9:
            slot = 0
        elif slot < 0:
            slot = 9

        self.selected_slot = slot
        self.update_slots()

    def update_slots(self):
        self.inventory_Surface.blit(self.inv_bar, (0, 0))
        self.inventory_Surface.blit(
            self.selected_slot_img, (self.inventory_offset+((self.selected_slot_img.get_width() - 2) * (self.selected_slot)), self.inventory_offset))
        for item in self.inventory.slots:
            if self.inventory.slots[item] is not None:
                self.inventory_Surface.blit(self.inventory.slots[item].item_image, (
                    self.inventory_offset+((self.selected_slot_img.get_width() - 2) * (item)) + 4, self.inventory_offset + 4))

                text = TextDisplay(
                    str(self.inventory.slots[item].count), 12, (255, 255, 255))

                self.inventory_Surface.blit(text.draw(), (
                    self.inventory_offset+((self.selected_slot_img.get_width() - 2) * (item)) + 4, self.inventory_offset + 4))


class CraftingHUD:
    def __init__(self, player) -> None:
        self.inventory_offset = 12
        self.inventory_offsetx = 8
        self.opened = False

        self.box = Box((500, 300), parent=self)
        self.box.position = (Config.screen.get_width() // 2 - self.box.Surface.get_width() // 2,
                             Config.screen.get_height() // 2 - self.box.Surface.get_height() // 2)

        self.player = player
        self.inventory = player.inventory

        self.slot_img = scale_image(pygame.image.load(
            Config.images["inventory_slot"]), 2)

        self.slots = {x: self.slot_img.copy() for x in range(9)}
        self.result = self.slot_img.copy()
        self.arrow = pygame.image.load(Config.images["arrow"])
        self.inv_bar = scale_image(pygame.image.load(
            Config.images["inventory_bar"]), 2)

        self.hud_surface = pygame.Surface(
            (self.slot_img.get_width() * 6, self.slot_img.get_height() * 3), pygame.SRCALPHA)

        self.craftable_items = Config.data["craft_items"]
        self.craftable_items_img = {
            x: load_item_image(x) for x in self.craftable_items}
        self.ingredients = {x: [(y, Config.data["craft_items"][x][y]) for y in Config.data["craft_items"][x]]
                            for x in self.craftable_items}

        self.craftable_items_hud = self.inv_bar.copy()

        self.selected_slot_img = scale_image(pygame.image.load(
            Config.images["inventory_slot_selected"]), 2)

        self.to_craft = None

        self.callback = None

        self.update()
        self.update_craftable_items()

    def add_event_to_box(self, event):
        self.box.set_close_event(event)

    def empty_slots(self):
        self.slots = {x: self.slot_img.copy() for x in range(9)}
        self.result = self.slot_img.copy()

    def toggle(self):
        self.box.opened = not self.box.opened
        self.update_craftable_items()
        self.update()

    def draw(self, screen: pygame.Surface):
        self.box.position = (screen.get_width() // 2 - self.box.size[0] // 2,
                             screen.get_height()//2 - self.box.size[1]//2)

        self.box.opened = True

        screen.blit(self.hud_surface, (screen.get_width() // 2 - self.hud_surface.get_width() // 2,
                                       screen.get_height()//2 - self.hud_surface.get_height()//2))

        screen.blit(self.craftable_items_hud, (screen.get_width() // 2 - self.craftable_items_hud.get_width() // 2,
                                               screen.get_height()//2 - self.craftable_items_hud.get_height()//2 - 100))
        self.mouse_on_item(pygame.mouse.get_pos(), screen)
        self.craft(screen)

    def update(self):

        for slot in self.slots:
            self.hud_surface.blit(
                self.slots[slot], (slot % 3 * self.slot_img.get_width(), slot // 3 * self.slot_img.get_height()))

        self.hud_surface.blit(
            self.result, (self.slot_img.get_width() * 5, self.slot_img.get_height()))

        self.hud_surface.blit(

            self.arrow, (int(self.slot_img.get_width() * 3.1), self.slot_img.get_height()))

        self.box.reset_and_add(self.hud_surface, (self.box.size[0] // 2 - self.hud_surface.get_width() // 2,
                                                  self.box.size[1] // 2 - self.hud_surface.get_height() // 2))

        self.box.add_element(self.craftable_items_hud, (self.box.size[0] // 2 - self.craftable_items_hud.get_width() // 2,
                                                        self.box.size[1] // 2 - self.craftable_items_hud.get_height() // 2 - 100))

    def update_craftable_items(self):
        self.craftable_items_hud = self.inv_bar.copy()
        for item in self.craftable_items:

            self.craftable_items_hud.blit(self.craftable_items_img[item], (self.inventory_offset + 4 + (
                (self.craftable_items_img[item].get_width(
                ) - 2 + self.inventory_offsetx)
                * ([*self.craftable_items.keys()].index(item))), self.inventory_offset + 4))

    def mouse_on_item(self, mouse_pos, screen):
        mouse_pos = (mouse_pos[0] - (screen.get_width() // 2 - self.box.Surface.get_width() // 2
                                     + (self.box.size[0] // 2 - self.craftable_items_hud.get_width() // 2)),
                     mouse_pos[1] - (screen.get_height()//2 - self.box.Surface.get_height()//2
                                     + (self.box.size[1] // 2 - self.craftable_items_hud.get_height() // 2 - 100)))

        for item in self.craftable_items:
            self.update_craftable_items()
            item_coords = (self.inventory_offset + 4 + (
                (self.craftable_items_img[item].get_width(
                ) - 2 + self.inventory_offsetx)
                * ([*self.craftable_items.keys()].index(item))), self.inventory_offset + 4)

            if (mouse_pos[0] >= item_coords[0]
                and mouse_pos[0] <= item_coords[0] + self.craftable_items_img[item].get_width()
                and mouse_pos[1] >= item_coords[1]
                    and mouse_pos[1] <= item_coords[1] + self.craftable_items_img[item].get_height()):

                self.craftable_items_hud.blit(
                    self.selected_slot_img,
                    (self.inventory_offset + (
                        (self.craftable_items_img[item].get_width()
                         - 2 + self.inventory_offsetx
                         )
                        * (
                            [*self.craftable_items.keys()].index(item)
                        )
                    ),
                        self.inventory_offset
                    )
                )

                Config.set_cursor_style(pygame.SYSTEM_CURSOR_HAND)

                self.update()
                if pygame.mouse.get_pressed()[0]:
                    self.empty_slots()
                    self.to_craft = None

                    self.set_crafint_table(item)
                return
        self.update_craftable_items()

    def set_crafint_table(self, item):

        for index, ingredient in enumerate(self.ingredients[item]):
            self.slots[index].blit(load_item_image(ingredient[0]), (4, 4))
            text = TextDisplay(
                str(ingredient[1]), 12, (255, 255, 255))
            self.slots[index].blit(text.draw(), (4, 4))

        self.to_craft = item
        self.result.blit(self.craftable_items_img[item], (4, 4))

        self.update()

    def craft(self, screen):
        mouse_pos = pygame.mouse.get_pos()
        mouse_pos = (mouse_pos[0] - (screen.get_width() // 2 - self.hud_surface.get_width() // 2),
                     mouse_pos[1] - (screen.get_height()//2 - self.hud_surface.get_height()//2))

        if (mouse_pos[0] >= self.slot_img.get_width() * 5
            and mouse_pos[0] <= self.slot_img.get_width() * 5 + self.slot_img.get_width()
            and mouse_pos[1] >= self.slot_img.get_height()
                and mouse_pos[1] <= self.slot_img.get_height() + self.slot_img.get_height()):

            Config.cursor_style = pygame.SYSTEM_CURSOR_HAND
            if pygame.mouse.get_pressed()[0] and self.to_craft:

                for ingredient in self.ingredients[self.to_craft]:
                    found = False
                    for item_index in self.inventory.slots:

                        if self.inventory.slots[item_index] and self.inventory.slots[item_index].type == ingredient[0]:
                            found = True
                            if self.inventory.slots[item_index].count < ingredient[1]:
                                return
                    if not found:
                        return

                for ingredient in self.ingredients[self.to_craft]:
                    self.inventory.subtract_item(
                        ingredient[0], ingredient[1])
                item_type = Config.data['item_types'][self.to_craft]

                match item_type:
                    case 'Weapon':
                        try:
                            damage = Config.data["damage"][self.to_craft]
                        except KeyError:
                            damage = 3
                        self.inventory.add_item_to_stack(Weapon(
                            self.craftable_items_img[self.to_craft],
                            self.to_craft,
                            Config.data['items_stack_size'][self.to_craft],
                            damage=damage,
                            durability=Config.data["durabilities"][self.to_craft]))

                    case 'Tool':
                        self.inventory.add_item_to_stack(Tool(
                            self.craftable_items_img[self.to_craft],
                            self.to_craft, Config.data['items_stack_size'][self.to_craft],
                            Config.data["durabilities"][self.to_craft]))
                    case 'Food':
                        self.inventory.add_item_to_stack(Food(
                            self.craftable_items_img[self.to_craft], self.to_craft,
                            Config.data['items_stack_size'][self.to_craft],
                            Config.data["food_health_bonus"][self.to_craft][0],
                            Config.data["food_health_bonus"][self.to_craft][1]))
                    case 'Material':
                        self.inventory.add_item_to_stack(Material(
                            self.craftable_items_img[self.to_craft],
                            self.to_craft,
                            Config.data['items_stack_size'][self.to_craft]))
                self.empty_slots()
                self.player.inventory_hud.update_slots()
                self.to_craft = None
                self.update()
                self.update_craftable_items()
                return

    @property
    def Surface(self):
        self.update_craftable_items()
        self.mouse_on_item(pygame.mouse.get_pos(), Config.screen)
        self.craft(Config.screen)

        self.box()

        return self.box



import json


class Config:
    data = {}
    images = {}

    screen = None
    map_layer = None
    game_map = None
    all_characters = []

    items = {
        "FOOD": [],
        "WOOD": [],
        "STICK": [],
        "STONE": [],
        "WOOD_SWORD": [],
        "STONE_SWROD": [],
        "PICKAXE": [],
        "AXE": [],
        "TORCH": [],


    }

    cursor_style = None

    @staticmethod
    def set_cursor_style(style):
        if not style:
            Config.cursor_style = style
            return

    @staticmethod
    def load(path: str):
        with open(path, 'r', encoding="UTF-8") as f:
            Config.data = json.load(f)

    @staticmethod
    def load_image_locations(path: str):
        with open(path, 'r', encoding="UTF-8") as f:
            Config.images = json.load(f)

    @staticmethod
    def remove_from_items(item_type, item):

        Config.items[item_type].pop(Config.items[item_type].index(item))



from setuptools import setup

setup(
    name="CatRoyale",
    version="1.18",
    description="A cat themed battle royale game",
    author="Tarsoly Barnabás",
    author_email="tarsoly.barnabas2002@gmail.com",
    url="https://github.com/Iseroo/Python-kotprog",
    packages=["cat_royale"],
    install_requires=["pygame", "webcolors", "pillow", "ordered_set"],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'CatRoyale = cat_royale.main:main_func'
        ]
    },

)



import random
import pygame
from pygame.locals import K_q, QUIT, KEYDOWN, K_ESCAPE, K_SPACE, K_e, K_c, K_f, K_m, K_LEFT, K_RIGHT, K_UP, K_DOWN, K_a, K_d, K_s, K_w
import sys
import webcolors
from cat_royale.classes.functions import scale_image
from cat_royale.classes.game_map import Block, GameMap
from cat_royale.classes.health_bars import HealthBar
from cat_royale.classes.map_reader import read_map_image, get_map_sprite_image
from cat_royale.classes.message_service import MessageService
from cat_royale.classes.item import Food, Material, ITEM, MAPCOLOR
from cat_royale.classes.config import Config
from cat_royale.classes.character import Character, Enemy
from cat_royale.classes.text_display import PlayerInfoText, TextDisplay
from cat_royale.classes.box import Box
from cat_royale.classes.event_stack import EventStack, Event, WindowStack

pygame.init()


class Game:

    def __init__(self):
        self.screen = pygame.display.set_mode(
            (Config.data["screen_size"]["width"], Config.data["screen_size"]["height"]))
        Config.screen = self.screen
        self.loading_text = TextDisplay(
            "Loading...", 30, (255, 255, 255))
        self.loadingScreen = pygame.image.load(Config.images["loading_screen"])
        self.screen.blit(self.loadingScreen, (self.screen.get_width()//2 -
                                              self.loadingScreen.get_width()//2, self.screen.get_height()//2 - self.loadingScreen.get_height()//2))
        self.screen.blit(self.loading_text.Surface, (self.screen.get_width()//2 - self.loading_text.Surface.get_width() //
                         2, self.screen.get_height()//2 - self.loading_text.Surface.get_height()//2 - 200))
        pygame.display.flip()
        EventStack.push(Event("close", self.close))

        self.clock = pygame.time.Clock()
        self.running = True
        self.craftHud_toggle = False

        self.img_size, self.png = read_map_image(
            Config.images["level00"])

        self.screen_layer = pygame.Surface(
            (self.img_size[0]*40, self.img_size[1]*40))

        self.map_layer = pygame.Surface(
            (self.img_size[0]*40, self.img_size[1]*40))
        Config.map_layer = self.map_layer
        self.camera = pygame.Surface(
            (self.img_size[0]*40, self.img_size[1]*40))
        self.make_map()
        self.camera_pos = (0, 0)
        self.camera_speed = 5
        self.update_camera()

        self.items = {}
        self.load_items()

        self.character = Character()
        Config.all_characters.append(self.character)

        self.enemies = [Enemy() for _ in range(100)]
        Config.all_characters.extend(self.enemies)
        for x in self.enemies:
            x.position = self.game_map.get_block_by_indexes(
                (random.randint(0, 80), random.randint(0, 80))).coords

        self.health_bar = HealthBar(
            (self.screen.get_width() // 2, self.screen.get_height()-80))

        self.player_info_text_display = PlayerInfoText(
            (self.character.get_position()[0], self.character.get_position()[1]-20))

        self.onblock = None

        self.game_over = False

        self.game_over_text = TextDisplay("Game Over", 24, (255, 0, 0))
        self.win_text = TextDisplay(
            "Winner winner cat dinner!", 24, (0, 255, 0))
        self.game_over_dialog = Box(
            (self.screen.get_width(), self.screen.get_height()), close_callback=self.close)
        self.game_over_dialog.add_element(self.game_over_text.Surface, (self.game_over_dialog.Surface.get_width(
        )//2-self.game_over_text.Surface.get_width()//2, self.game_over_dialog.Surface.get_height()//2-self.game_over_text.Surface.get_height()//2))

        self.minimap_img = pygame.image.load(Config.images["level00"])
        self.minimap = Box(
            (200, 200), close_button=False)

        self.minimap.position = (self.screen.get_width()-200,
                                 self.screen.get_height()-200)
        self.minimap.opened = True

    def determine_end(self):
        if self.character.hp <= 0:
            self.game_over = True

        if len([x for x in Config.all_characters if x.hp > 0]) == 1 and self.character.hp > 0:
            self.game_over = True
            self.game_over_dialog.reset_and_add(self.win_text.Surface, (self.game_over_dialog.Surface.get_width(
            )//2-self.win_text.Surface.get_width()//2, self.game_over_dialog.Surface.get_height()//2-self.win_text.Surface.get_height()//2))

    def close(self):
        self.running = False
        pygame.quit()
        sys.exit()

    def run(self):
        while self.running:

            Config.cursor_style = None
            self.clock.tick(60)
            self.move_camera()
            self.move_character()
            pygame.display.set_caption("FPS: " + str(self.clock.get_fps()))
            self.draw()

            self.handle_events()
            self.enemy_ai_updates()
            if Config.cursor_style:
                pygame.mouse.set_cursor(Config.cursor_style)
            else:
                pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_ARROW)
            self.update()

    def handle_events(self):

        for event in pygame.event.get():
            if event.type == QUIT:
                self.running = False
                pygame.quit()
                sys.exit()
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    EventStack.call_and_pop("close")

                if event.key == K_q:
                    dropped_item = self.character.inventory.remove_item_from_slot(
                        self.character.inventory_hud.selected_slot)
                    self.character.inventory_hud.update_slots()
                    if self.character.onblock and dropped_item:

                        self.character.onblock.add_item(dropped_item)
                        self.character.onblock.draw(self.map_layer)
                if event.key == K_f:
                    self.character.pickup()

                if event.key == K_c:
                    self.character.use_slot()

                if event.key == K_m:
                    self.minimap.opened = not self.minimap.opened
                if event.key == K_e:
                    self.character.crafting_hud.toggle()
                if event.key == K_SPACE:
                    self.character.do_attack = True
                    self.character.use_weapon_to_attack()
            if event.type == pygame.MOUSEWHEEL:
                if event.y > 0:
                    self.character.inventory_hud.select_slot(
                        self.character.inventory_hud.selected_slot + 1)
                elif event.y < 0:
                    self.character.inventory_hud.select_slot(
                        self.character.inventory_hud.selected_slot - 1)

        EventStack.find_and_call("Box_close_mouse_on")
        self.message_service_subscribe()

    def minimap_player(self):
        if self.character.onblock:

            copy = self.minimap_img.copy()
            pygame.draw.rect(copy, (0, 0, 255),
                             (*self.character.onblock.indexes, 2, 2))
            for enemy in self.enemies:
                if enemy.onblock and enemy.hp > 0:
                    pygame.draw.rect(copy, (0, 255, 255),
                                     (*enemy.onblock.indexes, 2, 2))
            copy = scale_image(copy, 1.5)

            self.minimap.Surface.blit(copy, (25, 25))

    def enemy_ai_updates(self):

        for enemy in self.enemies:
            enemy.ai()

    def update(self):
        self.character.set_onblock(self.onblock_for_character(self.character))
        for enemy in self.enemies:
            enemy.set_onblock(self.onblock_for_character(enemy))

        self.player_info_text_display.set_coords(
            (self.character.get_position()[0], self.character.get_position()[1]-20))

        self.health_bar.update(*self.character.get_health())

        if self.character.get_health()[0] <= 0:
            self.game_over = True
            self.character.position = (
                self.screen.get_width()//2-self.character.next_sprite.get_width()//2, self.screen.get_height()//2-100)

        pygame.display.flip()

    def onblock_for_character(self, character):
        return self.game_map.on_block_check(character.get_position())

    def draw(self):
        self.determine_end()
        if self.game_over:
            self.screen.blit(self.game_over_dialog.Surface, (0, 0))
            self.character.draw(self.screen)
            return
        self.screen.fill(MAPCOLOR.GRASS.rgb(MAPCOLOR.GRASS.value))
        self.screen_layer.blit(self.map_layer, (0, 0))
        self.character.draw(self.screen_layer)
        for enemy in self.enemies:
            enemy.draw(self.screen_layer)
        self.player_info_text_display.draw(self.screen_layer)
        self.screen.blit(self.screen_layer, self.camera_pos)

        self.character.inventory_hud.draw(self.screen)

        self.health_bar.draw(self.screen)
        if self.minimap.opened:
            self.minimap_player()
        for dialog in WindowStack.stack:
            if dialog.opened:
                self.screen.blit(dialog.Surface, dialog.position)

        if self.character.crafting_hud.box.opened:

            self.screen.blit(self.character.crafting_hud.Surface(),
                             self.character.crafting_hud.Surface.position)
        remaining_player_text = TextDisplay(
            f"{len([x for x in Config.all_characters if x.hp > 0])} player left.", 12, (255, 255, 255))
        self.screen.blit(remaining_player_text.Surface, (0, 0))

    def message_service_subscribe(self):
        message = MessageService.next()
        if message:
            color = (249, 113, 50) if message["severity"] == "warning" else (
                255, 0, 0) if message["severity"] == "error" else (255, 255, 255)
            text = TextDisplay(
                message["text"], 12, color)
            try:

                duration = message["duration"]
            except KeyError:
                duration = 100
            self.player_info_text_display.add(
                text, duration)

    def make_map(self):
        self.game_map = GameMap()
        Config.game_map = self.game_map
        sprite_sheet = pygame.image.load(
            Config.images["items"]).convert_alpha()

        for x in range(0, self.img_size[0]):
            for y in range(0, self.img_size[1]):
                mapcolor = MAPCOLOR(
                    webcolors.rgb_to_hex(self.png[x, y]).upper())
                block = Block((x*Block.size, y*Block.size),
                              mapcolor)
                self.game_map.add_block(
                    block)

                if mapcolor != MAPCOLOR.GRASS:
                    if mapcolor == MAPCOLOR.CARROT:
                        item = Food(get_map_sprite_image(
                            sprite_sheet, ITEM[mapcolor.name].value), mapcolor.name, Config.data["items_stack_size"][mapcolor.name],
                            Config.data["food_health_bonus"]["CARROT"][0], Config.data["food_health_bonus"]["CARROT"][1])
                        block.add_item(item)

                    elif mapcolor == MAPCOLOR.BERRY:
                        block_type = random.choice(
                            [ITEM["APPLE"], ITEM["BERRY"], ITEM["MUSHROOM"], ITEM["BERRY"]])
                        item = Food(get_map_sprite_image(
                            sprite_sheet, block_type.value), mapcolor.name, Config.data["items_stack_size"][mapcolor.name],
                            Config.data["food_health_bonus"][block_type.name][0], Config.data["food_health_bonus"][block_type.name][1])
                        block.add_item(item)
                    else:
                        item = Material(get_map_sprite_image(
                            sprite_sheet, ITEM[mapcolor.name].value), mapcolor.name, Config.data["items_stack_size"][mapcolor.name])
                        block.add_item(item)

    def move_camera(self, keys=None, speed=1):
        keys = pygame.key.get_pressed() if not keys else keys

        if keys[K_LEFT]:
            self.camera_pos = (
                self.camera_pos[0] + (self.camera_speed + (1 * speed)), self.camera_pos[1])
            self.check_camera_pos()
        if keys[K_RIGHT]:
            self.camera_pos = (
                self.camera_pos[0] - (self.camera_speed + (1 * speed)), self.camera_pos[1])
            self.check_camera_pos()
        if keys[K_UP]:
            self.camera_pos = (
                self.camera_pos[0], self.camera_pos[1] + (self.camera_speed + (1 * speed)))
            self.check_camera_pos()
        if keys[K_DOWN]:
            self.camera_pos = (
                self.camera_pos[0], self.camera_pos[1] - (self.camera_speed + (1 * speed)))
            self.check_camera_pos()

    def check_camera_pos(self):

        if self.camera_pos[0] > 0 + Config.data["camera_offset"]["x"]:
            self.camera_pos = (
                0 + Config.data["camera_offset"]["x"], self.camera_pos[1])
        if self.camera_pos[0] < -self.camera.get_width() + Config.data["screen_size"]["width"] - Config.data["camera_offset"]["x"]:
            self.camera_pos = (-self.camera.get_width() +
                               Config.data["screen_size"]["width"] - Config.data["camera_offset"]["x"], self.camera_pos[1])
        if self.camera_pos[1] > 0 + Config.data["camera_offset"]["y"]:
            self.camera_pos = (
                self.camera_pos[0], 0 + Config.data["camera_offset"]["y"])
        if self.camera_pos[1] < -self.camera.get_height() + Config.data["screen_size"]["height"] - Config.data["camera_offset"]["y"]:
            self.camera_pos = (self.camera_pos[0], -self.camera.get_height() +
                               Config.data["screen_size"]["height"] - Config.data["camera_offset"]["y"])

    def move_character(self):

        keys = pygame.key.get_pressed()

        if keys[K_a]:
            self.character.move("left")
        if keys[K_d]:
            self.character.move("right")
        if keys[K_w]:
            self.character.move("up")
        if keys[K_s]:
            self.character.move("down")

        keys = {K_LEFT: False, K_RIGHT: False, K_UP: False, K_DOWN: False}

        character_pos = (
            self.character.position[0] + self.camera_pos[0] +
            Config.data["camera_offset"]["x"],
            self.character.position[1] + self.camera_pos[1] +
            Config.data["camera_offset"]["y"])

        if character_pos[0] < 0 + Config.data["camera_offset"]["x"]:

            keys[K_LEFT] = True
        if character_pos[0] > Config.data["screen_size"]["width"] - Config.data["camera_offset"]["x"]:

            keys[K_RIGHT] = True
        if character_pos[1] < 0 + Config.data["camera_offset"]["y"]:

            keys[K_UP] = True
        if character_pos[1] > Config.data["screen_size"]["height"] - Config.data["camera_offset"]["y"]:

            keys[K_DOWN] = True
        self.move_camera(keys, speed=100)

    def update_camera(self):
        self.game_map.draw(self.map_layer)

    def load_items(self):
        sprite_sheet = pygame.image.load(
            Config.images["items"]).convert_alpha()
        for item in ITEM:
            self.items[item.name] = get_map_sprite_image(
                sprite_sheet, ITEM[item.name].value)


def main_func():

    Config.load("./config/config.json")
    Config.load_image_locations("./assets/image_locations.json")

    game = Game()

    game.run()




from PIL import Image
import pygame


def read_map_image(map_image):

    map_image = Image.open(map_image)
    map_image = map_image.convert('RGB')
    res = map_image.load()
    return (map_image.size, res)


def get_map_sprite_image(sprite_sheet: pygame.Surface, index: tuple, sprite_size=32):

    sprite = pygame.Surface((sprite_size, sprite_size), pygame.SRCALPHA)
    sprite.blit(sprite_sheet, (0, 0),
                (index[1] * sprite_size, index[0] * sprite_size, sprite_size, sprite_size))

    return sprite



import pygame
from cat_royale.classes.text_display import TextDisplay


class HealthBar:
    def __init__(self, position, hp=100, hunger=100) -> None:
        self._hp = hp
        self._hunger = hunger
        self.position = position
        self.bar_widht = 170  # px
        self.bar_height = 5  # px
        self.bar_gap = 30  # px

        self.hp_surface = pygame.Surface((self.bar_widht, self.bar_height))
        self.hunger_surface = pygame.Surface((self.bar_widht, self.bar_height))

        self.hp_text = TextDisplay("HP", 12, (255, 255, 255)).draw()

        self.hunger_text = TextDisplay("Hunger", 12, (255, 255, 255)).draw()

        self.make_bar_surfaces()

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp):
        if hp >= 0 or hp <= 100:
            self._hp = hp
            self.make_bar_surfaces()

    @property
    def hunger(self):
        return self._hunger

    @hunger.setter
    def hunger(self, hunger):
        if hunger >= 0 or hunger <= 100:
            self._hunger = hunger
            self.make_bar_surfaces()

    def update(self, hp, hunger):
        self.hp = hp
        self.hunger = hunger

    def make_bar_surfaces(self):
        self.hp_surface.fill((100, 0, 0))
        hp_rect = pygame.Rect(
            (0, 0), (self.bar_widht * self.hp / 100, self.bar_height))
        pygame.draw.rect(self.hp_surface, (255, 0, 0), hp_rect)

        self.hunger_surface.fill((0, 0, 100))
        hunger_rect = pygame.Rect(
            (0, 0), (self.bar_widht * self.hunger / 100, self.bar_height))

        pygame.draw.rect(self.hunger_surface, (0, 0, 255), hunger_rect)

    def draw(self, screen):
        position = (
            self.position[0] - ((self.bar_widht * 2 + self.bar_gap)//2), self.position[1])
        screen.blit(self.hp_surface, position)
        screen.blit(self.hunger_surface,
                    (position[0] + self.hp_surface.get_width() + self.bar_gap, position[1]))
        screen.blit(self.hp_text, (position[0], position[1] - 15))
        screen.blit(self.hunger_text,
                    (position[0] + 200, position[1] - 15))
